from http import HTTPStatus
from typing import List

from pydantic import parse_obj_as

from fiddler.libs.http_client import RequestClient
from fiddler.utils import logging
from fiddler.v2.schema.baseline import Baseline
from fiddler.v2.utils.exceptions import handle_api_error_response
from fiddler.v2.utils.response_handler import (
    APIResponseHandler,
    PaginatedResponseHandler,
)

logger = logging.getLogger(__name__)


class BaselineMixin:
    client: RequestClient
    organization_name: str

    @handle_api_error_response
    def get_baselines(
        self, project_name: str, model_name: str = None
    ) -> List[Baseline]:
        """Get list of all Baselines at project or model level

        :params project_name: project to list baselines from
        :params model_name: [Optional] model within project to list baselines from
        :returns: List containing Baseline objects
        """
        response = self.client.get(
            url='baselines/',
            params={
                'organization_name': self.organization_name,
                'project_name': project_name,
                'model_name': model_name,
            },
        )
        _, items = PaginatedResponseHandler(response).get_pagination_details_and_items()
        return parse_obj_as(List[Baseline], items)

    @handle_api_error_response
    def get_baseline(self, project_name: str, baseline_name: str) -> Baseline:
        """Get the details of a Baseline.

        :params project_name: The project to which the Baseline belongs to
        :params baseline_name: The Baseline name of which you need the details
        :returns: Baseline object which contains the details
        """
        response = self.client.get(
            url=f'baselines/{self.organization_name}:{project_name}:{baseline_name}',
        )
        response_handler = APIResponseHandler(response)
        return Baseline.deserialize(response_handler)

    @handle_api_error_response
    def add_baseline(
        self,
        project_name: str,
        name: str,
        type: str,
        model_association: str = None,
        dataset_name: str = None,
        model_name: str = None,
        start_time: int = None,
        end_time: int = None,
        offset: int = None,
        window_size: int = None,
    ) -> Baseline:
        """Function to add a Baseline to fiddler for monitoring

        :param project_name: project name where the Baseline will be added
        :type project_name: string
        :param name: name of the Baseline
        :type name: string
        :param type: type (DATASET, RELATIVE, STATIC) of the Baseline
        :type type: string
        :param model_association: (optional) model to associate baseline to
        :type model_association: string
        :param dataset_name: (optional) dataset to be used as baseline
        :type dataset_name: string
        :param model_name: (optional) model to be used as baseline
        :type model_name: string
        :param start_time: (optional) milliseconds since epoch to be used as start time for STATIC baseline
        :type start_time: int
        :param end_time: (optional) milliseconds since epoch to be used as end time for STATIC baseline
        :type end_time: int
        :param offset: (optional) seconds from current time to be used for RELATIVE baseline
        :type offset: int
        :param window_size: (optional) width of window in seconds to be used for RELATIVE baseline
        :type window_size: int

        :return: Baseline object which contains the Baseline details
        """

        request_body = Baseline(
            organization_name=self.organization_name,
            project_name=project_name,
            name=name,
            type=type,
            model_association=model_association,
            dataset_name=dataset_name,
            model_name=model_name,
            start_time=start_time,
            end_time=end_time,
            offset=offset,
            window_size=window_size,
        ).dict()

        if 'id' in request_body:
            request_body.pop('id')

        response = self.client.post(
            url='baselines/',
            data=request_body,
        )

        if response.status_code == HTTPStatus.OK:
            logger.info(f'{name} setup successful')
            return Baseline.deserialize(APIResponseHandler(response))

    @handle_api_error_response
    def delete_baseline(self, project_name: str, baseline_name: str) -> None:
        """Delete a Baseline

        :params project_name: Project name to which the Baseline belongs to.
        :params Baseline_name: Baseline name to be deleted

        :returns: None
        """
        response = self.client.delete(
            url=f'baselines/{self.organization_name}:{project_name}:{baseline_name}',
        )
        if response.status_code == HTTPStatus.OK:
            logger.info(f'{baseline_name} delete request received.')
        else:
            # @TODO: Handle non 200 status response
            logger.info('Delete unsuccessful')

    @handle_api_error_response
    def attach_baseline(
        self,
        project_name: str,
        name: str,
        model: str,
    ) -> None:
        """Function to add a Baseline to fiddler for monitoring

        :param project_name: project name where the Baseline will be added
        :type project_name: string
        :param name: name of the Baseline
        :type name: string
        :param model_association: (optional) model to associate baseline to
        :type model_association: string
        """

        response = self.client.post(
            url=f'baselines/{self.organization_name}:{project_name}:{name}:{model}',
            data={},
        )

        if APIResponseHandler(response).get_status_code() != HTTPStatus.OK:
            logger.info(f'Baseline {name} attachment to model {model} failed')
        else:
            logger.info(f'Baseline {name} successfully attached to model {model}')

    @handle_api_error_response
    def detach_baseline(
        self,
        project_name: str,
        name: str,
        model: str,
    ) -> None:
        """Function to remove baseline from model

        :param project_name: project name where the Baseline will be added
        :type project_name: string
        :param name: name of the Baseline
        :type name: string
        :param model: (optional) model to associate baseline to
        :type model: string
        """

        response = self.client.delete(
            url=f'baselines/{self.organization_name}:{project_name}:{name}:{model}',
            data={},
        )

        if APIResponseHandler(response).get_status_code() != HTTPStatus.OK:
            logger.info(f'Baseline {name} detachment from model {model} failed')
        else:
            logger.info(f'Baseline {name} successfully detached from model {model}')



from typing import Optional

from fiddler.v2.schema.base import BaseDataSchema


class Baseline(BaseDataSchema):
    id: Optional[int] = None
    name: str
    project_name: Optional[str]
    organization_name: Optional[str]
    type: Optional[str]
    model_association: Optional[str]

    dataset_name: Optional[str] = None
    model_name: Optional[str] = None
    start_time: Optional[int] = None
    end_time: Optional[int] = None
    offset: Optional[int] = None
    window_size: Optional[int] = None



from pydantic import BaseModel
from typing import Dict, Optional
from semver import VersionInfo


class Version(VersionInfo):

    @classmethod
    def __get_validators__(cls):
        """Return a list of validator methods for pydantic models."""
        yield cls.parse


class ServerInfo(BaseModel):
    features: Dict
    server_version: Optional[Version]



import pytest

from fiddler import FiddlerApi
from pytest_mock import MockFixture
from fiddler.v2.schema.server_info import Version


def test_client_v1_creation_fail():
    with pytest.raises(ValueError) as e:
        client_v1 = FiddlerApi('', '', '')


def test_client_v2_creation_fail():
    with pytest.raises(ValueError) as e:
        client_v2 = FiddlerApi('', '', '', version=2)


def test_get_server_info_without_server_version(mocker: MockFixture):
    mocker.patch('fiddler.connection.Connection.check_connection', return_value='OK')

    supported_features = {
        'features': {
            'enable_schema_creation': True,
            'authorization': True,
            'fairness': False,
            'dev_cluster': False,
        },
        'supported_client_version': '22.10.0',
        'enable_fiddler_v2': False,
    }
    mock_get_supported_features = mocker.patch('fiddler.FiddlerApi._get_supported_features')
    mock_get_supported_features.return_value = supported_features
    client = FiddlerApi('https://test.fiddler.ai', 'test', 'foo-token', version=1)

    assert client.v2.server_info.server_version is None
    assert client.v2.server_info.features == supported_features["features"]

    client = FiddlerApi('https://test.fiddler.ai', 'test', 'foo-token', version=2)

    assert client.v2.server_info.server_version is None
    assert client.v2.server_info.features == supported_features['features']


def test_get_server_info_with_server_version(mocker: MockFixture):
    mocker.patch('fiddler.connection.Connection.check_connection', return_value='OK')

    supported_features = {
        'features': {
            'enable_schema_creation': True,
            'authorization': True,
            'fairness': False,
            'dev_cluster': False,
        },
        'supported_client_version': '22.10.0',
        'enable_fiddler_v2': False,
        'server_version': "22.10.0"
    }
    mock_get_supported_features = mocker.patch('fiddler.FiddlerApi._get_supported_features')
    mock_get_supported_features.return_value = supported_features
    client = FiddlerApi('https://test.fiddler.ai', 'test', 'foo-token', version=1)

    assert client.v2.server_info.server_version == Version.parse(supported_features['server_version'])
    assert client.v2.server_info.features == supported_features["features"]

    client = FiddlerApi('https://test.fiddler.ai', 'test', 'foo-token', version=2)

    assert client.v2.server_info.server_version == Version.parse(supported_features['server_version'])
    assert client.v2.server_info.features == supported_features['features']



import re
import numpy as np
import pandas as pd
from fiddler import gem


class IGTextAttributionsTF2Keras:
    """
    Helper class for project attribution method when computing IG for Text data with Keras TF2.
    """
    def __init__(self, input_df, output_cols):
        """
        :param input_df: pandas dataframe for a single observation
        :param output_cols: list of output column names
        """
        self.input_df = input_df
        self.output_cols = output_cols

    def text_to_tokens_keras(self, tokenizer, max_seq_length, feature_label):
        """
        Helper function to convert text to tokens with keras TF2.

        :param tokenizer: keras tokenizer used during training
        :param max_seq_length: max sequence length used during training
        :param feature_label: str. Name of the text feature input
        :return: word tokens
        """
        from tensorflow.keras.preprocessing.sequence import pad_sequences

        unpadded_tokens = [
            tokenizer.texts_to_sequences([x])[0]
            for x in self.input_df[feature_label].values
        ]

        padded_tokens = pad_sequences(
            unpadded_tokens, max_seq_length, padding='post', truncating='post'
        )

        word_tokens = tokenizer.sequences_to_texts([[x] for x in padded_tokens[0]])
        return word_tokens

    def text_attributions(self, tokenizer, word_tokens, word_attributions, feature_label):
        """
        Helper function to define segments works and attributions.

        :param tokenizer: keras tokenizer used during training
        :param word_tokens: word tokens. Could be the output of the text_to_tokens_keras method
        :param word_attributions: associated word attributions
        :param feature_label: str. Name of the text feature input
        :return: final segments and attributions
        """
        segments = re.split(
            r'([ ' + tokenizer.filters + '])',
            self.input_df.iloc[0][feature_label],
        )
        i = 0
        final_attributions = []
        final_segments = []
        for segment in segments:
            if segment != '':  # dump empty tokens
                final_segments.append(segment)
                seg_low = segment.lower()
                if len(word_tokens) > i and seg_low == word_tokens[i]:
                    final_attributions.append(word_attributions[i])
                    i += 1
                else:
                    final_attributions.append(0)
        return final_segments, final_attributions

    def get_attribution_for_output(self, att, word_tokens, tokenizer, embedding_name, feature_label):
        """
        Helper function to get attributions for a given output.

        :param att: dictionary of attributions for the given output
        :param word_tokens: word tokens
        :param tokenizer: keras tokenizer used during training
        :param embedding_name: str. Name of the embedding layer in the model
        :param feature_label: str. Name of the text feature input
        :return:
        """
        # Note - summing over attributions in the embedding direction
        word_attributions = np.sum(att[embedding_name][-len(word_tokens):],
                                   axis=1)
        final_segments, final_attributions = self.text_attributions(tokenizer, word_tokens, word_attributions,
                                                                    feature_label)
        gem_text = gem.GEMText(feature_name=feature_label,
                               text_segments=final_segments,
                               text_attributions=final_attributions)

        return gem_text

    def get_project_attribution(self, attributions, tokenizer, word_tokens, embedding_name, feature_label):
        """
        Helper method to get project attributions when model has a single text input feature.

        :param attributions: list of IG attributions. Each element of the list corresponds to an output.
        :param tokenizer: tokenizer used during training
        :param word_tokens: word tokens
        :param embedding_name: str. Name of the embedding layer in the model
        :param feature_label: str. Name of the text feature input
        :return:
        """
        explanations_by_output = {}

        if isinstance(feature_label, list):
            if len(feature_label) == 1:
                feature_label = feature_label[0]
            else:
                raise ValueError("Your model has multiple inputs. You cannot use this helper. "
                                 "Please implement project_attributions accordingly. "
                                 "If you need some help, contact Fiddler.")
        if isinstance(embedding_name, list):
            if len(embedding_name) == 1:
                embedding_name = embedding_name[0]
            else:
                raise ValueError("Your model has multiple embeddings. You cannot use this helper. "
                                 "Please implement project_attributions accordingly. "
                                 "If you need some help, contact Fiddler.")

        for output_field_index, att in enumerate(attributions):
            gem_text = self.get_attribution_for_output(att, word_tokens, tokenizer, embedding_name, feature_label)
            gem_container = gem.GEMContainer(contents=[gem_text])
            explanations_by_output[self.output_cols[output_field_index]] = gem_container.render()

        return explanations_by_output


class IGTabularAttributions:
    """
    Helper class for project attribution method when computing IG for Tabular data.
    """

    def __init__(self, input_df, output_cols):
        """
        :param input_df: pandas dataframe for a single observation
        :param output_cols: list of output column names
        """
        self.input_df = input_df
        self.output_cols = output_cols

    def get_project_attribution(self, attributions, attr_input_names_mapping):
        """
        Helper method to get project attributions when input data is tabular.

        :param attributions: list of IG attributions. Each element of the list corresponds to an output.
        :param attr_input_names_mapping: dict that map attributable layer names to input feature names
        :return: dictionary explanations_by_output
        """
        explanations_by_output = {}
        if attr_input_names_mapping is None:
            raise ValueError('Parameter attr_input_names_mapping cannot be empty when IG is enabled.')
        for output_field_index, att in enumerate(attributions):
            explanations_by_output[self.output_cols[output_field_index]] = \
                get_tabular_attributions_for_output(self.input_df, att,
                                                    attr_input_names_mapping=attr_input_names_mapping)
        return explanations_by_output


class TreeShapAttributions:
    """
    Helper class for computing attributions with Tree SHAP for tree based model
    """

    def __init__(self, input_df, transformed_df, output_cols, model):
        """
        :param input_df: pandas dataframe for a single observation
        :param transformed_df: pandas dataframe for a single transformed observation
               (if any pre-processing steps are required before running the model on the data)
        :param output_cols: list of output column names
        """
        import shap

        self.explainer = shap.TreeExplainer(model)
        self.input_df = input_df
        self.transformed_df = transformed_df
        self.output_cols = output_cols

    def get_project_attribution(self, predictions_dict, transformed_names_mapping=None):
        """
        Helper method to get project attributions for Tree SHAP.

        :param predictions_dict: dictionary of predictions of the model for the given input
        :param transformed_names_mapping: dictionary mapping feature name to corresponding transformed columns
        :return: dictionaries explanations_by_output, extras_by_output
        """
        if not isinstance(predictions_dict, dict):
            raise ValueError("The parameter predictions_dict has to be a mapping dictionary between output col name "
                             "and prediction value for the given input.")
        if set(predictions_dict.keys()) != set(self.output_cols):
            raise ValueError("The parameter predictions_dict has to have output col names as keys.")

        shap_values = self.explainer.shap_values(self.transformed_df, check_additivity=False)
        untransformed_base_values = self.explainer.expected_value
        if isinstance(untransformed_base_values, float):
            untransformed_base_values = [untransformed_base_values]

        explanations_by_output = {}
        extras_by_output = {}

        if (transformed_names_mapping is None) and (set(self.input_df.columns) != set(self.transformed_df.columns)):
            raise ValueError("Your data has some pre-processing steps before being fed to the model. "
                             "Please provide a dictionary with key feature names and values list of corresponding "
                             "transformed feature names.")

        for ind, output_name in enumerate(self.output_cols):
            if len(self.output_cols) == 1:
                if len(shap_values) == 2:
                    # For binary classification models that predict proba for 0 and 1
                    attributions = shap_values[1][0]
                else:
                    attributions = shap_values[0]
            else:
                attributions = shap_values[ind][0]

            # Transform shap values from log-odds to probability space
            shap_values_transformed, base_value =\
                xgb_shap_transform_scale(attributions, model_prediction=predictions_dict[output_name],
                                         untransformed_base_value=untransformed_base_values[ind])

            if transformed_names_mapping is not None:
                # Map the transformed input names to shap values
                map_att_transformed = {name: shap_values_transformed[ind]
                                       for ind, name in enumerate(self.transformed_df.columns)}
                # Map the original input names to sum of shap values
                shap_values_transformed = {name: np.sum([map_att_transformed[col]
                                                         for col in transformed_names_mapping[name]])
                                           for name in transformed_names_mapping.keys()}
            explanations_by_output[output_name] = get_tabular_attributions_for_output(self.input_df,
                                                                                      shap_values_transformed)

            extras_by_output[output_name] = {'model_prediction': predictions_dict[output_name],
                                             'baseline_prediction': base_value}

        return explanations_by_output, extras_by_output


def get_tabular_attributions_for_output(input_df, attributions,
                                        attr_input_names_mapping=None):
    """
    Helper function to get attributions for a single output.

    :param input_df: pandas dataframe for a single observation
    :param attributions: dictionary of attributions with keys attributable input names if IG is enabled.
                         Otherwise, this is just a list of attributions.
    :param attr_input_names_mapping: Required only if IG is enabled.
    :return: explanations for a single output
    """
    dense_features = []

    if attr_input_names_mapping is None:
        if isinstance(attributions, (list, np.ndarray)):
            attributions = {'dense_inputs': attributions}
            attr_input_names_mapping = {'dense_inputs': list(input_df.columns)}
        elif isinstance(attributions, dict):
            attr_input_names_mapping = {'dense_inputs': list(attributions.keys())}
            attributions = {'dense_inputs': list(attributions.values())}
        else:
            raise ValueError(f"The argument attributions should be a list or a dictionary mapping."
                             f" It's currently of type: {type(attributions)}")

    for key in attributions.keys():
        if key not in attr_input_names_mapping:
            raise ValueError(f"The key {key} is missing in the attr_input_names_mapping dictionary.")
        for ind, feature_name in enumerate(attr_input_names_mapping[key]):
            attr_val = attributions[key][ind]
            if isinstance(attr_val, (list, np.ndarray)):
                attr_val = np.sum(attr_val)
            value = input_df[feature_name].iloc[0]
            if pd.isnull(value):
                value = None
            dense_features.append(
                gem.GEMSimple(
                    feature_name=feature_name,
                    value=value,
                    attribution=float(attr_val),
                )
            )
        # Attribute 0.0 for features not used in the model
        missing_features = set(input_df.columns) - set(attr_input_names_mapping[key])
        for feature_name in missing_features:
            value = input_df[feature_name].iloc[0]
            if pd.isnull(value):
                value = None
            dense_features.append(
                gem.GEMSimple(
                    feature_name=feature_name,
                    value=value,
                    attribution=float(0.0),
                )
            )
    gem_container = gem.GEMContainer(contents=dense_features)

    return gem_container.render()


def xgb_shap_transform_scale(shap_values, model_prediction, untransformed_base_value):
    base_value = (1 / (1 + np.exp(-untransformed_base_value)))

    # Computing the original_explanation_distance to construct the distance_coefficient later on
    original_explanation_distance = np.sum(shap_values)

    # Computing the distance between the model_prediction and the transformed base_value
    distance_to_explain = model_prediction - base_value

    if np.abs(distance_to_explain) <= 0.0001:
        # In that case, attributions are null
        shap_values_transformed = np.array([0.0] * len(shap_values))
    else:
        # The distance_coefficient is the ratio between both distances which will be used later on
        distance_coefficient = original_explanation_distance / distance_to_explain
        # Transforming the original shapley values to the new scale
        shap_values_transformed = shap_values / distance_coefficient

    return shap_values_transformed, base_value



from http import HTTPStatus
from typing import List, Optional
import os
from deepdiff import DeepDiff

from pydantic import parse_obj_as

from fiddler.libs.http_client import RequestClient
from fiddler.utils import logging
from fiddler.v2.schema.model import Model
from fiddler.v2.utils.exceptions import handle_api_error_response
from fiddler.v2.utils.response_handler import (
    APIResponseHandler,
    PaginatedResponseHandler,
)
from fiddler.v2.api.helpers import multipart_upload_model
from fiddler.core_objects import ModelInfo, ArtifactStatus
from fiddler.utils.general_checks import type_enforce
import yaml
from fiddler.model_info_validator import ModelInfoValidator
from pathlib import Path

logger = logging.getLogger(__name__)


class ModelMixin:
    client: RequestClient
    organization_name: str

    @handle_api_error_response
    def get_models(self, project_name: str) -> List[Model]:
        """
        Get list of all models belonging to a project

        :params project_name: The project for which you want to get the models
        :returns: List containing Model objects
        """
        response = self.client.get(
            url='models',
            params={
                'organization_name': self.organization_name,
                'project_name': project_name,
            },
        )
        _, items = PaginatedResponseHandler(response).get_pagination_details_and_items()
        return parse_obj_as(List[Model], items)

    @handle_api_error_response
    def get_model(self, project_name: str, model_name: str) -> Model:
        """
        Get the details of a model.

        :params project_name: The project to which the model belongs to
        :params model_name: The model name of which you need the details
        :returns: Model object which contains the details
        """
        response = self.client.get(
            url=f'models/{self.organization_name}:{project_name}:{model_name}',
        )
        response_handler = APIResponseHandler(response)
        return Model.deserialize(response_handler)

    @handle_api_error_response
    def add_model(
        self,
        project_name: str,
        model_name: str,
        info: Optional[dict] = None,
        model_type: Optional[str] = None,
        framework: Optional[str] = None,
        requirements: Optional[str] = None,
    ) -> Model:
        """
        Function to add a model to fiddler for monitoring

        :param project_name: project name where the model will be added
        :type project_name: string
        :param model_name: name of the model
        :type model_name: string
        :param info: model related information passed as dictionary from user
        :type info: dictionary
        :param model_type: the type of model e.g. REGRESSION, BINARY CLASSIFICATION
        :type model_type: string
        :return: Model object which contains the model details
        """
        # @TODO: file_list required while setting up model?
        # @TODO: Handle model_deployment_id?
        request_body = Model(
            organization_name=self.organization_name,
            project_name=project_name,
            name=model_name,
            info=info,
            model_type=model_type,
            framework=framework,
            requirements=requirements,
        ).dict()

        # remove id key from POST request, as it will be created by the backend
        if 'id' in request_body:
            request_body.pop('id')

        response = self.client.post(
            url='models',
            data=request_body,
        )
        logger.info(f'{model_name} setup successful')

        return Model.deserialize(APIResponseHandler(response))

    @handle_api_error_response
    def delete_model(self, project_name: str, model_name: str) -> None:
        """
        Delete a model

        :params project_name: Project name to which the model belongs to.
        :params model_name: Model name to be deleted

        :returns: None
        """
        response = self.client.delete(
            url=f'models/{self.organization_name}:{project_name}:{model_name}',
        )
        if response.status_code == HTTPStatus.OK:
            logger.info(f'{model_name} delete request received.')
        else:
            # @TODO: Handle non 200 status response
            logger.info('Delete unsuccessful')

    @handle_api_error_response
    def upload_model_package(
        self,
        project_name: str,
        model_name: str,
        artifact_path: Path,
    ) -> Model:
        artifact_path = type_enforce('artifact_path', artifact_path, Path)

        if not artifact_path.is_dir():
            raise ValueError(f'Not a valid model dir: {artifact_path}')

        yaml_file = artifact_path / 'model.yaml'
        if not yaml_file.is_file():
            raise ValueError(f'Model yaml not found {yaml_file}')

        with yaml_file.open() as f:
            model_info = ModelInfo.from_dict(yaml.safe_load(f))

        miv = ModelInfoValidator(model_info, None)
        miv.validate_categoricals(modify=True)

        response = multipart_upload_model(self.client, self.organization_name, project_name, model_name, artifact_path)

        if response.get('status') == 'SUCCESS':
            file_list = []
            for _, _, files in os.walk(artifact_path):
                for name in files:
                    abs_path = artifact_path / name
                    file_info = {'name': abs_path.name, 'size': abs_path.stat().st_size, 'modified': abs_path.stat().st_mtime}
                    file_list.append(file_info)

            # 2. Create postgres and CH entry
            request_body = Model(
            organization_name=self.organization_name,
            project_name=project_name,
            name=model_name,
            info=model_info.to_dict(),
            file_list={'files':file_list},
            framework='',
            model_type='',
            requirements='',
            ).dict()

            if 'id' in request_body:
                request_body.pop('id')

            response = self.client.post(
                url='models',
                data=request_body,
                params={'blob_uploaded': True}
            )
            logger.info(f'{model_name} setup successful')

            return Model.deserialize(APIResponseHandler(response))


    @handle_api_error_response
    def add_model_surrogate(
        self,
        project_name: str,
        dataset_name: str,
        model_name: str,
        info: dict,
    ) -> Model:
        model_info = ModelInfo.from_dict(info)
        if model_info.artifact_status is None:
            model_info.artifact_status = ArtifactStatus.SURROGATE
        if model_info.artifact_status != ArtifactStatus.SURROGATE:
            raise ValueError(f'Please set model_info.artifact_status to {ArtifactStatus.SURROGATE}')
    
        request_body = Model(
            organization_name=self.organization_name,
            project_name=project_name,
            name=model_name,
            info=model_info.to_dict(),
        ).dict(include={'organization_name', 'project_name', 'info', 'name'})

        request_body.update({'dataset_name': dataset_name})

        response = self.client.post(
           url='model-surrogate',
           data=request_body
        )

        logger.info(f'{model_name} surrogate created successfully!')

        return Model.deserialize(APIResponseHandler(response))

    @handle_api_error_response
    def update_model(
        self,
        project_name: str,
        model_name: str,
        artifact_path: Path,
    ) -> Model:
        artifact_path = type_enforce('artifact_path', artifact_path, Path)

        if not artifact_path.is_dir():
            raise ValueError(f'Not a valid model dir: {artifact_path}')

        yaml_file = artifact_path / 'model.yaml'
        if not yaml_file.is_file():
            raise ValueError(f'Model yaml not found {yaml_file}')

        with yaml_file.open() as f:
            model_info = ModelInfo.from_dict(yaml.safe_load(f))

        if len(model_info.datasets) < 1:
            raise ValueError('Unable to find dataset in model.yaml')

        if len(model_info.datasets) > 1:
            raise ValueError('More than one dataset specified in model.yaml')

        remote_model_info = ModelInfo.from_dict(self.get_model(project_name, model_name).info)

        # Model info should remain exactly the same, expect the following fields can be changed:
        # custom_explanation_names, preferred_explanation_method, display_name, description, framework, algorithm
        # and model_deployment_params
        ddiff = DeepDiff(remote_model_info, model_info, ignore_order=True,
                         exclude_paths=['root.preferred_explanation_method', 'root.custom_explanation_names',
                                        'root.display_name', 'root.description', 'root.framework', 'root.algorithm',
                                        'root.model_deployment_params',  'root.artifact_status'])

        if len(ddiff) > 0:
            raise ValueError(
                f'remote model info, does not match '
                f'local model info: {ddiff}. Updating those fields in model info is '
                f'not currently supported'
            )

        response = multipart_upload_model(self.client, self.organization_name, project_name, model_name, artifact_path, is_update=True)
        if response.get('status') == 'SUCCESS':
            file_list = []
            for _, _, files in os.walk(artifact_path):
                for name in files:
                    abs_path = artifact_path / name
                    file_info = {'name': abs_path.name, 'size': abs_path.stat().st_size, 'modified': abs_path.stat().st_mtime}
                    file_list.append(file_info)

            # 2. Create postgres and CH entry
            request_body = Model(
            organization_name=self.organization_name,
            project_name=project_name,
            name=model_name,
            info=model_info.to_dict(),
            file_list={'files':file_list}
            ).dict(include={'name', 'project_name', 'organization_name', 'info', 'file_list'})

            if 'id' in request_body:
                request_body.pop('id')

            response = self.client.patch(
                url=f'models/{self.organization_name}:{project_name}:{model_name}',
                data=request_body,
            )
            logger.info(f'{model_name} updated successful')

            return Model.deserialize(APIResponseHandler(response))
